@@ -28,6 +28,7 @@ from .ml.pipeline import MLCarefreePipeline
 from .cv.models.diffusion import ldm as _ldm
 from .cv.models.diffusion import ldm_vq as _ldm_vq
 from .cv.models.diffusion import ldm_sd as _ldm_sd
+from .cv.models.diffusion import ldm_sd_v2 as _ldm_sd_v2
 from .cv.models.diffusion import ldm_sr as _ldm_sr
 from .cv.models.diffusion import ldm_sd_anime as _ldm_sd_anime
 from .cv.models.diffusion import ldm_semantic as _ldm_semantic
@@ -1080,6 +1081,16 @@ def chinese_clip_model(pretrained: bool = True, **kwargs: Any) -> CLIP:
     return DLZoo.load_model(model, pretrained=pretrained, **kwargs)
 
 
+def open_clip_ViT_H_14(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
+    model = "multimodal/clip.open_clip_ViT_H_14"
+    return DLZoo.load_pipeline(model, pretrained=pretrained, **kwargs)
+
+
+def open_clip_ViT_H_14_model(pretrained: bool = True, **kwargs: Any) -> CLIP:
+    model = "multimodal/clip.open_clip_ViT_H_14"
+    return DLZoo.load_model(model, pretrained=pretrained, **kwargs)
+
+
 def clip_vqgan_aligner(**kwargs: Any) -> DLPipeline:
     return DLZoo.load_pipeline("multimodal/clip_vqgan_aligner", **kwargs)
 
@@ -1348,6 +1359,14 @@ def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
     return _ldm_sd_anime(pretrained)
 
 
+def ldm_sd_inpainting(pretrained: bool = True) -> DLPipeline:
+    return _ldm_sd_inpainting(pretrained)
+
+
+def ldm_sd_v2(pretrained: bool = True) -> DLPipeline:
+    return _ldm_sd_v2(pretrained)
+
+
 def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
     return _ldm_celeba_hq(pretrained)
 
@@ -1356,10 +1375,6 @@ def ldm_inpainting(pretrained: bool = True) -> DLPipeline:
     return _ldm_inpainting(pretrained)
 
 
-def ldm_sd_inpainting(pretrained: bool = True) -> DLPipeline:
-    return _ldm_sd_inpainting(pretrained)
-
-
 def ldm_sr(pretrained: bool = True) -> DLPipeline:
     return _ldm_sr(pretrained)
 
