@@ -1032,13 +1032,17 @@ def ldm_sd_version(version: str, pretrained: bool = True) -> DLPipeline:
 
 
 def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
-    return ldm_sd(pretrained, download_name="ldm_sd_anime_nai")
+    return ldm_sd_version("anime_nai", pretrained)
 
 
 def ldm_sd_inpainting(pretrained: bool = True, **kw: Any) -> DLPipeline:
     return _ldm("diffusion/ldm.sd_inpainting", 64, 9, 4, pretrained=pretrained, **kw)
 
 
+def ldm_sd_v2(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
+    return _ldm("diffusion/ldm.sd_v2", 64, 4, 4, pretrained=pretrained, **kwargs)
+
+
 def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
     return ldm_vq(
         pretrained=pretrained,
