@@ -9,6 +9,7 @@ from torch import Tensor
 from typing import Any
 from typing import Dict
 from typing import List
+from typing import Tuple
 from typing import Optional
 from typing import Protocol
 
@@ -31,8 +32,24 @@ def to_d(image: Tensor, sigma: Tensor, denoised: Tensor) -> Tensor:
     return (image - denoised) / append_dims(sigma, image.ndim)
 
 
+def get_ancestral_step(
+    sigma_from: float,
+    sigma_to: float,
+    eta: float = 1.0,
+) -> Tuple[float, float]:
+    if not eta:
+        return sigma_to, 0.0
+    sigma_up = min(
+        sigma_to,
+        eta
+        * (sigma_to**2 * (sigma_from**2 - sigma_to**2) / sigma_from**2) ** 0.5,
+    )
+    sigma_down = (sigma_to**2 - sigma_up**2) ** 0.5
+    return sigma_down, sigma_up
+
+
 class IGetDenoised(Protocol):
-    def __call__(self, img: Tensor) -> Tensor:
+    def __call__(self, img: Tensor, sigma: Tensor) -> Tensor:
         pass
 
 
@@ -126,16 +143,18 @@ class KSamplerMixin(ISampler, UncondSamplerMixin, metaclass=ABCMeta):
         if step == total_step - 1:
             return image
 
-        def get_denoised(img: Tensor) -> Tensor:
-            sigmas = self.sigmas[step] * image.new_ones([image.shape[0]])
-            ts = self._sigma_to_t(sigmas, quantize)
+        def get_denoised(img: Tensor, sigma: Tensor) -> Tensor:
+            sigma = sigma * s_in
+            ts = self._sigma_to_t(sigma, quantize)
             c_in = append_dims(
-                1.0 / (sigmas**2 + self.sigma_data**2) ** 0.5, img.ndim
+                1.0 / (sigma**2 + self.sigma_data**2) ** 0.5,
+                img.ndim,
             )
-            c_out = append_dims(sigmas, img.ndim)
+            c_out = append_dims(sigma, img.ndim)
             eps = self._uncond_denoise(img * c_in, ts, cond)
             return img - eps * c_out
 
+        s_in = image.new_ones([image.shape[0]])
         return self.sample_step_core(
             image,
             cond,
@@ -217,7 +236,7 @@ class KLMSSampler(KSamplerMixin):
         **kwargs: Any,
     ) -> Tensor:
         order = 4
-        denoised = get_denoised(image)
+        denoised = get_denoised(image, self.sigmas[step])
         d = to_d(image, self.sigmas[step], denoised)
         self.ds.append(d)
         if len(self.ds) > order:
@@ -245,6 +264,120 @@ class KLMSSampler(KSamplerMixin):
         self.sigmas_numpy = self.sigmas.detach().cpu().numpy()
 
 
+@ISampler.register("k_euler")
+class KEulerSampler(KSamplerMixin):
+    def sample_step_core(
+        self,
+        image: Tensor,
+        cond: Optional[cond_type],
+        step: int,
+        total_step: int,
+        get_denoised: IGetDenoised,
+        *,
+        quantize: bool,
+        unconditional_cond: Optional[Any],
+        unconditional_guidance_scale: float,
+        **kwargs: Any,
+    ) -> Tensor:
+        s_churn = 0.0
+        s_tmin = 0.0
+        s_tmax = float("inf")
+        s_noise = 1.0
+        sigma = self.sigmas[step]
+        if not s_tmin <= sigma.item() <= s_tmax:
+            gamma = 0.0
+        else:
+            gamma = min(s_churn / (total_step - 1), 2**0.5 - 1)
+        eps = torch.randn_like(image) * s_noise
+        sigma_hat = sigma * (gamma + 1)
+        if gamma > 0:
+            image = image + eps * (sigma_hat**2 - sigma**2) ** 0.5
+        denoised = get_denoised(image, sigma_hat)
+        d = to_d(image, sigma_hat, denoised)
+        dt = self.sigmas[step + 1] - sigma_hat
+        image = image + d * dt
+        return image
+
+
+@ISampler.register("k_euler_a")
+class KEulerAncestralSampler(KSamplerMixin):
+    def sample_step_core(
+        self,
+        image: Tensor,
+        cond: Optional[cond_type],
+        step: int,
+        total_step: int,
+        get_denoised: IGetDenoised,
+        *,
+        quantize: bool,
+        unconditional_cond: Optional[Any],
+        unconditional_guidance_scale: float,
+        **kwargs: Any,
+    ) -> Tensor:
+        eta = 1.0
+        s_noise = 1.0
+        sigma = self.sigmas[step]
+        next_sigma = self.sigmas[step + 1]
+        denoised = get_denoised(image, sigma)
+        sigma_down, sigma_up = get_ancestral_step(
+            sigma.item(),
+            next_sigma.item(),
+            eta=eta,
+        )
+        d = to_d(image, sigma, denoised)
+        dt = sigma_down - sigma
+        image = image + d * dt
+        if next_sigma.item() > 0:
+            image = image + torch.randn_like(image) * s_noise * sigma_up
+        return image
+
+
+@ISampler.register("k_heun")
+class KHeunSampler(KSamplerMixin):
+    def sample_step_core(
+        self,
+        image: Tensor,
+        cond: Optional[cond_type],
+        step: int,
+        total_step: int,
+        get_denoised: IGetDenoised,
+        *,
+        quantize: bool,
+        unconditional_cond: Optional[Any],
+        unconditional_guidance_scale: float,
+        **kwargs: Any,
+    ) -> Tensor:
+        s_churn = 0.0
+        s_tmin = 0.0
+        s_tmax = float("inf")
+        s_noise = 1.0
+        sigma = self.sigmas[step]
+        next_sigma = self.sigmas[step + 1]
+        if not s_tmin <= sigma.item() <= s_tmax:
+            gamma = 0.0
+        else:
+            gamma = min(s_churn / (total_step - 1), 2**0.5 - 1)
+        eps = torch.randn_like(image) * s_noise
+        sigma_hat = sigma * (gamma + 1)
+        if gamma > 0:
+            image = image + eps * (sigma_hat**2 - sigma**2) ** 0.5
+        denoised = get_denoised(image, sigma_hat)
+        d = to_d(image, sigma_hat, denoised)
+        dt = next_sigma - sigma_hat
+        if next_sigma.item() == 0:
+            image = image + d * dt
+        else:
+            x_2 = image + d * dt
+            denoised_2 = get_denoised(x_2, next_sigma)
+            d_2 = to_d(x_2, next_sigma, denoised_2)
+            d_prime = (d + d_2) / 2
+            image = image + d_prime * dt
+        return image
+
+
 __all__ = [
     "KLMSSampler",
+    "KEulerSampler",
+    "KEulerAncestralSampler",
+    "KHeunSampler",
 ]
