@@ -4,20 +4,43 @@ from torch import Tensor
 from typing import Any
 from typing import Dict
 from typing import Optional
-from cftool.misc import check_requires
+from cftool.types import tensor_dict_type
 
-from .encoder import EncoderBase
-from .encoder import Encoder1DBase
-from .decoder import DecoderBase
-from .decoder import Decoder1DBase
-from ...types import tensor_dict_type
+from .encoder import make_encoder
+from .encoder import IEncoder
+from .encoder import EncoderMixin
+from .decoder import make_decoder
+from .decoder import IDecoder
 from ...misc.toolkit import auto_num_layers
 
 
-class EncoderDecoder(nn.Module):
+class EncoderDecoderMixin:
+    encoder: IEncoder
+    decoder: IDecoder
+
+    def resize(self, net: Tensor) -> Tensor:
+        return self.decoder.resize(net)
+
+    def encode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
+        return self.encoder.encode(batch, **kwargs)
+
+    def decode(
+        self,
+        batch: tensor_dict_type,
+        *,
+        resize: bool = True,
+        **kwargs: Any,
+    ) -> Tensor:
+        net = self.decoder.decode(batch, **kwargs)
+        if resize:
+            net = self.resize(net)
+        return net
+
+
+class EncoderDecoder(nn.Module, EncoderDecoderMixin):
     def __init__(
         self,
-        is_1d: int,
+        is_1d: bool,
         in_channels: int,
         out_channels: Optional[int] = None,
         target_downsample: Optional[int] = None,
@@ -47,19 +70,12 @@ class EncoderDecoder(nn.Module):
                 )
             args = img_size, min_size, target_downsample
             num_downsample = auto_num_layers(*args, use_stride=encoder == "vanilla")
-        num_upsample = num_upsample or num_downsample
         if latent_resolution is None and img_size is None:
             raise ValueError(
                 "either `img_size` or `latent_resolution` should be provided "
                 "in `VanillaVAEBase` (when `latent_resolution` is not provided, "
                 "it will be inferred automatically with `img_size`)"
             )
-        # check num_downsample requirement
-        self.encoder_base = (Encoder1DBase if is_1d else EncoderBase).get(encoder)  # type: ignore
-        self.decoder_base = (Decoder1DBase if is_1d else DecoderBase).get(decoder)  # type: ignore
-        requires_num_downsample = check_requires(self.encoder_base, "num_downsample")
-        if not requires_num_downsample:
-            num_downsample = None
         # properties
         self.is_1d = is_1d
         self.latent = latent
@@ -73,38 +89,42 @@ class EncoderDecoder(nn.Module):
         # encoder
         if encoder_config is None:
             encoder_config = {}
-        if check_requires(self.encoder_base, "img_size"):
-            encoder_config["img_size"] = img_size
-        if requires_num_downsample:
-            encoder_config["num_downsample"] = num_downsample
-        encoder_config["in_channels"] = in_channels
+        encoder_config.setdefault("img_size", img_size)
+        encoder_config.setdefault("in_channels", in_channels)
+        encoder_config.setdefault("num_downsample", num_downsample)
         encoder_config[self.latent_key] = latent
-        self.encoder = self.encoder_base.make(encoder, config=encoder_config)
-        if not is_1d and img_size is not None:
+        self.encoder = make_encoder(encoder, encoder_config, is_1d=is_1d)  # type: ignore
+        if isinstance(self.encoder, EncoderMixin) and img_size is not None:
             latent_resolution = self.encoder.latent_resolution(img_size)
         # decoder
         if decoder_config is None:
             decoder_config = {}
+        decoder_config.setdefault("img_size", img_size)
+        decoder_config.setdefault("num_classes", num_classes)
+        decoder_config.setdefault("out_channels", out_channels or in_channels)
+        decoder_config.setdefault("num_upsample", num_upsample)
+        decoder_config.setdefault("latent_resolution", latent_resolution)
         decoder_config[self.latent_key] = self.decoder_latent
-        decoder_config["img_size"] = img_size
-        decoder_config["num_classes"] = num_classes
-        decoder_config["out_channels"] = out_channels or in_channels
-        if num_upsample is not None and not decoder.startswith("style"):
-            decoder_config["num_upsample"] = num_upsample
-        if latent_resolution is not None:
-            decoder_config["latent_resolution"] = latent_resolution
-        self.decoder = self.decoder_base.make(decoder, config=decoder_config)
+        self.decoder = make_decoder(decoder, decoder_config, is_1d=is_1d)
         self.latent_resolution = self.decoder.latent_resolution
 
-    def resize(self, net: Tensor) -> Tensor:
-        return self.decoder.resize(net)
 
-    def encode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
-        return self.encoder.encode(batch, **kwargs)
-
-    def decode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
-        net = self.decoder.decode(batch, **kwargs)
-        return self.resize(net)
+class PureEncoderDecoder(nn.Module, EncoderDecoderMixin):
+    def __init__(
+        self,
+        *,
+        is_1d: bool,
+        encoder: str = "vanilla",
+        decoder: str = "vanilla",
+        encoder_config: Optional[Dict[str, Any]] = None,
+        decoder_config: Optional[Dict[str, Any]] = None,
+    ):
+        super().__init__()
+        self.encoder = make_encoder(encoder, encoder_config or {}, is_1d=is_1d)
+        self.decoder = make_decoder(decoder, decoder_config or {}, is_1d=is_1d)
 
 
-__all__ = ["EncoderDecoder"]
+__all__ = [
+    "EncoderDecoder",
+    "PureEncoderDecoder",
+]
