@@ -1,11 +1,11 @@
 import json
 import torch
 
-from torch import nn
 from torch import Tensor
 from typing import Dict
 from typing import List
 from typing import Union
+from typing import Optional
 from cftool.misc import shallow_copy_dict
 from cftool.array import tensor_dict_type
 
@@ -51,7 +51,8 @@ def _convert_cond_stage(d: tensor_dict_type, md: tensor_dict_type) -> tensor_dic
         nd[key] = md.pop(key)
     prefix = "cond_stage_model.transformer.text_model.encoder.layers"
     m_prefix = "core.condition_model.m.text_transformer.encoder.mixing_blocks"
-    num_cond_layers = int(list(d)[-3].split(".")[-3]) + 1
+    cond_keys = [k for k in d if k.startswith("cond_stage_model")]
+    num_cond_layers = int(cond_keys[-3].split(".")[-3]) + 1
     for i in range(num_cond_layers):
         i_prefix = f"{prefix}.{i}"
         m_i_prefix = f"{m_prefix}.{i}"
@@ -248,19 +249,29 @@ def _convert_with_key_mapping(
 def _get_d(inp: Union[str, tensor_dict_type]) -> tensor_dict_type:
     if isinstance(inp, str):
         inp = torch.load(inp, map_location="cpu")
-        if "state_dict" in inp:
-            inp = inp["state_dict"]
-    return inp
+    if "state_dict" in inp:
+        inp = inp["state_dict"]
+    return shallow_copy_dict(inp)
 
 
 def convert(
     inp: Union[str, tensor_dict_type],
     api: DiffusionAPI,
     *,
+    vae_inp: Optional[Union[str, tensor_dict_type]] = None,
     load: bool = False,
     use_mapping: bool = True,
 ) -> tensor_dict_type:
     inp = _get_d(inp)
+    if vae_inp is not None:
+        vae_inp = _get_d(vae_inp)
+        num_injected = 0
+        for k, v in vae_inp.items():
+            inp_k = f"first_stage_model.{k}"
+            if inp_k in inp:
+                inp[inp_k] = v
+                num_injected += 1
+        print(f">> injected vae parameters: {num_injected}")
     with api.load_context() as wrapper:
         md = wrapper.state_dict()
         if not use_mapping:
